/***********************************************************************
 *                                                                     *
 *                               Decoder                               *
 *                                                                     *
 ***********************************************************************
 ***********************************************************************/
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <zlib.h>
#include <stdarg.h>
#include "gmap.h"
#include "bitmap.h"
#include "pngio.h"
#include "bmk.h"
#include "firead.h"

#define MAGIC_SIZE 8
static unsigned char MAGIC[8]= "IMGDOC01";

struct imdo
{
   // Header information
  unsigned int bmap_section_offset, bmap_section_size,
               gmap_section_offset, gmap_section_size,
               page_section_offset, page_section_size,
               bmk_section_offset, bmk_section_size;

  unsigned int *bmap_table;
  unsigned int n_bmap;
  // bitmaps are cached, so I'm going to keep them here.
  // They aren't too big or too many to keep in memory.
  gmap_t **bmap;

  // Graymaps are cached only if they are small enough.
  // Otherwise, they are unlikely to be repeated in
  // multiple pages. The threshold is given in the
  // initializer.
  unsigned int *gmap_table;
  unsigned int n_gmap;
  gmap_t **gmap;
  int gmap_cache_threshold;

  unsigned int *page_table;
  unsigned int n_page;

  bmk_t *bmk;
  int nobmk; // This flag is set when there is an error reading the
             // bookmarks or when there are no bookmarks in the file.

  FILE *f;
  char *error;
};

static int im_error(imdo_t *IM, const char *fmt, ...)
{
  char buf[2000];
  va_list ap;
  va_start(ap,fmt);
  vsnprintf(buf, sizeof(buf)-1, fmt, ap);
  va_end(ap);
  buf[sizeof(buf)-1]= 0;
  IM->error= strdup(buf);
  return 1;
}

static int set_fpos(imdo_t *IM, unsigned int pos)
{
  if (fseek(IM->f, pos, SEEK_SET))
    return im_error(IM,"can't seek in output file: %s\n",strerror(errno));
  return 0;
}

static int read_bytes(imdo_t *IM, void *dst, int len)
{
  int R;
  R= fread(dst,1,len,IM->f);
  if (R!=len)
    return im_error(IM,"error during read");
  return 0;
}

static int alloc_read(imdo_t *IM, void **dst, int len)
{
  void *K;
  *dst= NULL;
  K= malloc(len);
  if (read_bytes(IM, K,len)) { free(K); return 1; }
  *dst= K;
  return 0;
}


static unsigned int dint4(unsigned char *V)
{
  return V[0] | (V[1]<<8) | (V[2]<<16) | (V[3]<<24);
}

static unsigned int read_int4(imdo_t *IM, unsigned int *V)
{
  unsigned char B[4];
  if (read_bytes(IM,B,4)) return 1;
  *V= dint4(B); 
  return 0;
}

static int pos_and_size(imdo_t *IM, unsigned int pos, unsigned int *sz)
{
  if (set_fpos(IM, pos) || read_int4(IM,sz)) return 1;
  return 0;
}

static unsigned int dint2(unsigned char *K)
{
  return K[0] | (K[1]<<8);
}

static int read_hdr(imdo_t *IM)
{
  return set_fpos(IM, MAGIC_SIZE)  
      || read_int4(IM,&IM->bmap_section_size)
      || read_int4(IM,&IM->bmap_section_offset)
      || read_int4(IM,&IM->gmap_section_size)
      || read_int4(IM,&IM->gmap_section_offset)
      || read_int4(IM,&IM->page_section_size)
      || read_int4(IM,&IM->page_section_offset)
      || read_int4(IM,&IM->bmk_section_size)
      || read_int4(IM,&IM->bmk_section_offset);
}

static int read_bmap_table(imdo_t *IM)
{
  if (pos_and_size(IM, IM->bmap_section_offset, &IM->n_bmap)) return 1;
  if (alloc_read(IM, (void**) &IM->bmap_table, 4*IM->n_bmap)) return 1;
  IM->bmap= calloc(sizeof(IM->bmap[0]),IM->n_bmap);
  return 0;
}

static int read_gmap_table(imdo_t *IM)
{
  if (pos_and_size(IM, IM->gmap_section_offset, &IM->n_gmap)) return 1;
  if (alloc_read(IM, (void**) &IM->gmap_table, 4*IM->n_gmap)) return 1;
  IM->gmap= calloc(sizeof(IM->gmap[0]),IM->n_gmap);
  return 0;
}


static int read_bmap(imdo_t *IM, int num)
{
  unsigned int sz;
  unsigned char *b;
  if (IM->bmap[num]) return 0;
  if (pos_and_size(IM, IM->bmap_table[num],&sz)) return 1;
  if (alloc_read(IM, (void**) &b, sz)) return 1;
  IM->bmap[num]= malloc(sizeof(gmap_t));
  IM->bmap[num]->data= bitmap_to_graymap(b, &IM->bmap[num]->w,
                                            &IM->bmap[num]->h);
  free(b);
  return 0;
}


static int read_gmap(imdo_t *IM, int num)
{
  unsigned int sz;
  unsigned char *b;
  char *e;
  unsigned int W,H; unsigned char *D;

  if (IM->gmap[num]) return 0;
  if (pos_and_size(IM, IM->gmap_table[num], &sz)) return 1;
  if (alloc_read(IM, (void**) &b, sz)) return 1;

  IM->gmap[num]= malloc(sizeof(gmap_t));
  if ((e=decode_png_g8(b+4, sz, &W, &H, &D)))
  {
   W= 1; H=1; D= (unsigned char*) strdup("");
   free(e);
  }
  IM->gmap[num]->data= D;
  IM->gmap[num]->w= W;
  IM->gmap[num]->h= H; 
  free(b);
  return 0;
}

static int read_page_table(imdo_t *IM)
{
  if (pos_and_size(IM, IM->page_section_offset, &IM->n_page)) return 1;
  return alloc_read(IM, (void**) &IM->page_table, 4*IM->n_page);
}

static int read_page
    (imdo_t *IM, int pg, uint16_t *Rw, uint16_t *Rh,uint8_t *Rcolor,
              uint8_t **Rcmd,int *Rcmdlen, uint8_t **Rfreep)
{
  unsigned char *raw;
  unsigned char *dec;
  unsigned int rawlen;
  uLong decomp_sz, decR;
  
  if (pos_and_size(IM, IM->page_table[pg], &rawlen)) return 1;
  if (alloc_read(IM, (void**) &raw, rawlen)) return 1;

  decomp_sz= dint4(raw);
  *Rw= dint2(raw+4);
  *Rh= dint2(raw+6);
  *Rcolor= raw[8];

  if (decomp_sz==0) // uncompressed
  {
    *Rcmd= raw+9;
    *Rcmdlen= rawlen - 9;
    *Rfreep= raw;
    return 0;
  }

  dec= malloc(decomp_sz);
  decR= decomp_sz;

  if (uncompress(dec, &decR, raw+9, rawlen-9)!=Z_OK || (decR!=decomp_sz))
  {
    free(raw);
    *Rfreep= *Rcmd= (unsigned char*) strdup("");
    *Rcmdlen= 0;
    return 0;
  }

  free(raw);
  *Rfreep= *Rcmd= dec;
  *Rcmdlen= decomp_sz;
  return 0;
}


int imdo_render_page(imdo_t *IM,int pg, unsigned char **Rimg,
                     unsigned int *Rw,unsigned int *Rh)
{
  unsigned char *U,*end;
  unsigned char *img;
  gmap_t *S;
  int cX,cY, x,y, cW, cH, cc;
  uint16_t W,H;
  unsigned char bg;
  int no;
  unsigned char *CMD, *freep;
  int CMDL;

  if (IM->error) return 1;

  freep= NULL;
  img= NULL;

  if (read_page(IM, pg,&W,&H,&bg, &CMD, &CMDL, &freep)) return 1;

  end= CMD + CMDL;
  U= CMD;
  img= malloc(W*H);
  memset(img, bg, W*H);

  while(U<end)
  switch(U[0])
  {
  case 'R':
     U++;
     cX= dint2(U); U+=2;
     cY= dint2(U); U+=2;
     cW= dint2(U); U+=2;
     cH= dint2(U); U+=2;
     cc= U[0];     U++;
     for(y=0;y<cH;y++)
      for(x=0;x<cW;x++)
        img[(x+cX)+(y+cY)*W]= cc;
     break;
  case 'G': 
     U++;
     cX= dint2(U); U+= 2;
     cY= dint2(U); U+= 2;
     no= dint4(U); U+= 4;
     if (read_gmap(IM,no)) goto fail;
     S= IM->gmap[no];
     for(y=0;y<S->h;y++)
       for(x=0;x<S->w;x++)
         img[ (x+cX)+ (y+cY)*W ] = S->data[x+y*S->w];
     if (S->w*S->h>IM->gmap_cache_threshold)
     {
       free(S->data);
       free(S);
       IM->gmap[no]= NULL;
     }
     break;
  case 'B':
     U++;
     cX= dint2(U); U+= 2;
     cY= dint2(U); U+= 2;
     no= dint4(U); U+= 4;
     if (read_bmap(IM,no)) goto fail;
     S= IM->bmap[no];
     for(y=0;y<S->h;y++)
       for(x=0;x<S->w;x++)
         img[ (x+cX)+ (y+cY)*W ] &= S->data[x+y*S->w];
     break;
  }

  *Rimg= img; *Rw= W; *Rh= H;
  free(freep);
  return 0;
fail:
  if (freep) free(freep);
  if (img)   free(img);
  return 1;
}


int imdo_init(imdo_t **RIM,char *fn, int cache_thr)
{
  imdo_t *IM;
  char fmagic[MAGIC_SIZE];

  *RIM= IM= calloc(1,sizeof(*IM));
  IM->f= fopen(fn, "rb");
  if (!IM->f) return im_error(IM,"can't open input %s: %s",fn,strerror(errno));
  if (read_bytes(IM, fmagic, MAGIC_SIZE))
     return im_error(IM,"%s: can't read magic", fn);
  if (memcmp(fmagic,MAGIC, MAGIC_SIZE))
     return im_error(IM,"%s: bad magic", fn);

  IM->gmap_cache_threshold= cache_thr;
  
  return read_hdr(IM)
         || read_bmap_table(IM)
         || read_gmap_table(IM)
         || read_page_table(IM);
}

static void freenotnull(void *a) { if (a) free(a); }

static void free_gmap_array(gmap_t **A, int N)
{
  int i;
  if (!A) return ;
  for(i=0;i<N;i++)
    if (A[i])
    {
      freenotnull(A[i]->data);
      free(A[i]);
    }
  free(A);
}

static void free_bookmarks(bmk_t *B)
{
  if (!B) return ;
  if (B->children)
  {
    bmk_t *C,*N;
    for(C=B->children->next;C!=B->children;C=N)
    {
      N= C->next;
      free_bookmarks(C);
    }
    free(C);
  }
  if (B->title) free(B->title);
  free(B);
}

void imdo_close(imdo_t *IM)
{
  if (IM->f) fclose(IM->f);

  freenotnull(IM->page_table);
  freenotnull(IM->bmap_table);
  freenotnull(IM->gmap_table);

  free_gmap_array(IM->bmap, IM->n_bmap);
  free_gmap_array(IM->gmap, IM->n_gmap);

  free_bookmarks(IM->bmk);

  freenotnull(IM->error);
  memset(IM, 0, sizeof(*IM));
  free(IM);
}

char *imdo_error(imdo_t *IM)
{
  return IM->error;
}

int imdo_npages(imdo_t *IM, unsigned int *npage)
{
  *npage= IM->n_page;
  return 0;
}

bmk_t *imdo_bookmarks(imdo_t *IM)
{
  unsigned int dsiz;
  unsigned char *data;
  if (IM->bmk) return IM->bmk;
  if (IM->nobmk) return NULL;

  IM->nobmk= 1;
  if (IM->error || IM->bmk_section_size==0 ||
                   IM->bmk_section_offset==0) return NULL;
  if (pos_and_size(IM,IM->bmk_section_offset,&dsiz)) return NULL;
  if (alloc_read(IM, (void**) &data, dsiz)) return NULL;

  IM->bmk= bmk_decode(data, dsiz);
  free(data);
  if (IM->bmk) IM->nobmk= 0;
  return IM->bmk; 
}
