/***********************************************************************
 *                                                                     *
 *                               Merger                                *
 *                                                                     *
 ***********************************************************************
  The merger works as follows. It goes thru the input files in parallel.
  First, we seek to the bitmap area and make a merge operation just like
  in merge sort. We write the combined bitmap area to some place and
  continue with the graymaps. 
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdint.h>
#include <zlib.h>
#include "bmk.h"
#include "util-encode.c"
#include "util-encode-2.c"

#define TMPFILE "tmp.imd.p"

typedef struct
{
  uint32_t SS_bmap, SO_bmap, SS_gmap, SO_gmap,
           SS_page, SO_page, SS_bmk, SO_bmk;
  uint32_t n_bmap, *bmap_table, *bmap_rename;
  uint32_t n_gmap, *gmap_table, *gmap_rename;
  uint32_t n_page, *page_table;

  FILE *f;
  char *name;
} inp_t;

FILE *open_force(char *name, const char *mode)
{
  FILE *f;
  f= fopen(name,mode);
  if (!f) exit(fprintf(stderr,"can't open %s in %s mode: %s\n",
               name, mode, strerror(errno)));
  return f;
}

static void read_bytes(FILE *f, void *buf, int len)
{
  int R;
  R= fread(buf,1,len,f);
  if (R!=len)
  {
    if (feof(f)) exit(fprintf(stderr,"can't read from file: eof\n"));
    exit(fprintf(stderr,"can't read from file: %s\n", strerror(errno)));
  }
}

static unsigned int read_int4(inp_t *inp)
{
  unsigned char V[4];
  read_bytes(inp->f,V,4);
  return (V[0]) | (V[1]<<8) | (V[2]<<16) | (V[3]<<24);
}

static void write_int4(FILE *f,uint32_t V)
{
  unsigned char U[4];
 
  U[0]= V; U[1]= V>>8; U[2]= V>>16; U[3]= V>>24;
  write_bytes(f,U,4);
}

static int little_endian()
{
  union { uint32_t iv; unsigned char av[4]; } M;
  M.iv= 0xdeadbeef;
  if (M.av[0]==0xef) return 1;
                else return 0;
}

static void convert_int4_ltoh(unsigned int *K, unsigned int N)
{
  unsigned int i, V;
  unsigned char *Y;
  Y= (unsigned char*) K;
  for(i=0;i<N;i++)
  {
     V= Y[0] | (Y[1]<<8) | (Y[2]<<16) | (Y[3]<<24);
     K[i]= V;
     Y+= 4;
  }
}

static void convert_int4_htol(unsigned int *K, unsigned int N)
{
  unsigned int i, V;
  unsigned char *Y;
  Y= (unsigned char*) K;
  for(i=0;i<N;i++)
  {
     V= K[i];
     Y[0]= V;
     Y[1]= V>>8;
     Y[2]= V>>16;
     Y[3]= V>>24;
     Y+= 4;
  }
}

static void read_table(inp_t *inp, unsigned int offset,
                       unsigned int *Rn_items, unsigned int **Rtable)
{
  unsigned int n_items, *table;
  set_fpos(inp->f, offset);
  n_items= read_int4(inp);
  table= malloc(n_items*4);
  read_bytes(inp->f, table, n_items*4);
  if (!little_endian()) convert_int4_ltoh(table, n_items);
  *Rn_items= n_items;
  *Rtable= table;
}



static void read_header_and_tables(inp_t *inp)
{
  char fmagi[MAGIC_SIZE];

  read_bytes(inp->f, fmagi, MAGIC_SIZE);
  if (memcmp(fmagi, MAGIC, MAGIC_SIZE))
    exit(fprintf(stderr,"magic doesn't match for %s\n",inp->name));

  inp->SS_bmap= read_int4(inp);
  inp->SO_bmap= read_int4(inp);
  inp->SS_gmap= read_int4(inp);
  inp->SO_gmap= read_int4(inp);
  inp->SS_page= read_int4(inp);
  inp->SO_page= read_int4(inp);
  inp->SS_bmk=  read_int4(inp);
  inp->SO_bmk=  read_int4(inp);

  read_table( inp, inp->SO_bmap, &inp->n_bmap, &inp->bmap_table );
  read_table( inp, inp->SO_gmap, &inp->n_gmap, &inp->gmap_table );
  read_table( inp, inp->SO_page, &inp->n_page, &inp->page_table );
}


static void read_obj(inp_t *inp, uint32_t offset,
                     uint32_t *Rsiz, unsigned char **Rdata)
{
  unsigned char *data;
  unsigned int siz;

  set_fpos(inp->f, offset);
  siz= read_int4(inp);
  data= malloc(siz);
  read_bytes(inp->f, data, siz);
  *Rsiz= siz;
  *Rdata= data;
}

static uint32_t write_obj(FILE *f, uint32_t sz, unsigned char *D)
{
  write_int4(f, sz);
  write_bytes(f, D, sz);
  return sz+4;
}

static int compare_obj(uint32_t Asiz, unsigned char *Adata,
                       uint32_t Bsiz, unsigned char *Bdata)
{
  uint32_t i;
  for(i=0;i<Asiz && i<Bsiz;i++)
  {
    if (Adata[i]>Bdata[i]) return 1;
    if (Adata[i]<Bdata[i]) return -1;
  }
  if (Asiz==Bsiz) return 0;// the missing bytes are assumed to be zero
  if (Asiz>Bsiz) return 1; // this is the same logic used when creating
  return -1;               // the object (in objmap based on critbit). 
}

static void combine_objects
  (inp_t *inpA, uint32_t nA, uint32_t *tableA, uint32_t **RrenameA,
   inp_t *inpB, uint32_t nB, uint32_t *tableB, uint32_t **RrenameB,
   uint32_t *Rnobj, uint32_t *Roffset, uint32_t **Rtable)
{
  uint32_t pA, pB;
  unsigned char *Adata, *Bdata;
  uint32_t Asiz, Bsiz;
  uint32_t *renameA, *renameB;
  uint32_t offset, n_obj, *tableO;
  int R;

  FILE *ofile;

  ofile= open_force(TMPFILE, "wb");

  pA= 0; pB= 0;
  Adata= Bdata= NULL;
  renameA= calloc(nA,sizeof(uint32_t));
  renameB= calloc(nB,sizeof(uint32_t));
  tableO= calloc(nA+nB,sizeof(uint32_t));
  offset= 0; n_obj= 0;

  if (pA!=nA) read_obj(inpA, tableA[pA], &Asiz, &Adata);
  if (pB!=nB) read_obj(inpB, tableB[pB], &Bsiz, &Bdata);

  while(pA!=nA || pB!=nB)
  {
    if (pA==nA) R=1;
    else if (pB==nB) R= -1;
    else R= compare_obj(Asiz, Adata, Bsiz, Bdata);

    if (R<=0) renameA[pA]= n_obj;
    if (R>=0) renameB[pB]= n_obj;
    tableO[n_obj]= offset;
    n_obj++;
    if (R<=0) offset+= write_obj(ofile,Asiz, Adata);
    if (R>0)  offset+= write_obj(ofile,Bsiz, Bdata);

    if (R<=0) { pA++; if (pA!=nA) read_obj(inpA, tableA[pA], &Asiz, &Adata); }
    if (R>=0) { pB++; if (pB!=nB) read_obj(inpB, tableB[pB], &Bsiz, &Bdata); }
  }

  *RrenameA= renameA;
  *RrenameB= renameB;
  *Rnobj= n_obj;
  *Roffset= offset;
  *Rtable= tableO;

  if (Adata) free(Adata);
  if (Bdata) free(Bdata);

  fclose(ofile);
}

/* When this function is done, the table is no longer usable.
   This is because we convert the table to little-endian 
   format in the function. If the host is LE, it could still
   be used, but no bets. */
static void write_table(FILE *f, uint32_t n, uint32_t *table)
{
  write_int4(f, n);
  if (!little_endian()) convert_int4_htol(table, n);
  write_bytes(f, table, n*4);
}

static void copy_tmpfile(FILE *out)
{
  FILE *in;
  int R;
  unsigned char buffer[2000];
  in= open_force(TMPFILE,"rb");
  while((R=fread(buffer,1,sizeof(buffer),in)))
    write_bytes(out, buffer,R);
  fclose(in);
  unlink(TMPFILE);
}

static void combine_bitmaps(FILE *fO,inp_t *inpA,inp_t *inpB)
{
  uint32_t n_bmap, data_size, table_size, *bmap_table, i; 
  uint32_t bmap_SO, bmap_SS;

  combine_objects(inpA, inpA->n_bmap, inpA->bmap_table, &inpA->bmap_rename,
                  inpB, inpB->n_bmap, inpB->bmap_table, &inpB->bmap_rename,
                  &n_bmap, &data_size, &bmap_table);

  bmap_SO= HDR_TOTAL;
  table_size= n_bmap*4;
  for(i=0;i<n_bmap;i++)
    bmap_table[i] += bmap_SO + table_size + 4;

  set_fpos(fO, bmap_SO);
  write_table(fO, n_bmap, bmap_table);

  copy_tmpfile(fO);

  bmap_SS= 4 + table_size + data_size;

  set_fpos(fO, HDR_BMAP);
  write_int4(fO, bmap_SS);
  write_int4(fO, bmap_SO);

  free(bmap_table);
printf("input bitmaps= %d + %d= %d\n", inpA->n_bmap, inpB->n_bmap,
                                       inpA->n_bmap+inpB->n_bmap);
printf("output bitmaps= %d\n", n_bmap);
}

static void combine_graymaps(FILE *fO,inp_t *inpA,inp_t *inpB)
{
  uint32_t n_gmap, data_size, *gmap_table, i; 
  uint32_t gmap_SO, gmap_SS, table_size;

  combine_objects(inpA, inpA->n_gmap, inpA->gmap_table, &inpA->gmap_rename,
                  inpB, inpB->n_gmap, inpB->gmap_table, &inpB->gmap_rename,
                  &n_gmap, &data_size, &gmap_table);

  seek_end(fO);
  gmap_SO= get_fpos(fO);
  table_size= n_gmap*4;
  for(i=0;i<n_gmap;i++)
    gmap_table[i] += gmap_SO + table_size + 4;

  set_fpos(fO, gmap_SO);
  write_table(fO, n_gmap, gmap_table);

  copy_tmpfile(fO);

  gmap_SS= 4 + table_size + data_size;

  set_fpos(fO, HDR_GMAP);
  write_int4(fO, gmap_SS);
  write_int4(fO, gmap_SO);

  free(gmap_table);
printf("input graymaps= %d + %d= %d\n", inpA->n_gmap, inpB->n_gmap,
                                       inpA->n_gmap+inpB->n_gmap);
printf("output graymaps= %d\n", n_gmap);
}

static uint32_t dint4(unsigned char *D)
{
  return (D[0]) | (D[1]<<8) | (D[2]<<16) | (D[3]<<24);
}

static uint16_t dint2(unsigned char* D)
{
  return D[0] | (D[1]<<8);
}

static int Oint4(unsigned char *D, uint32_t V)
{
  D[0]= V; D[1]= V>>8; D[2]= V>>16; D[3]= V>>24;
  return 4;
}

static int Oint2(unsigned char *D, uint32_t V)
{
  D[0]= V; D[1]= V>>8; 
  return 2;
}

static int Oint1(unsigned char *D, uint32_t V)
{
  D[0]= V; 
  return 1;
}


static void read_page
    (inp_t *inp, int pg, uint16_t *Rw, uint16_t *Rh,uint8_t *Rcolor,
                uint8_t **Rcmd,int *Rcmdlen, uint8_t **Rfreep)
{
  unsigned char *raw;
  unsigned char *dec;
  unsigned int rawlen;
  uLong decomp_sz, decR;
  
  set_fpos(inp->f, inp->page_table[pg]);
  rawlen= read_int4(inp);
  raw= malloc(rawlen);
  read_bytes(inp->f, raw, rawlen);

  decomp_sz= dint4(raw);
  *Rw= dint2(raw+4);
  *Rh= dint2(raw+6);
  *Rcolor= raw[8];

  if (decomp_sz==0) // uncompressed
  {
    *Rcmd= raw+9;
    *Rcmdlen= rawlen - 9;
    *Rfreep= raw;
    return ;
  }

  dec= malloc(decomp_sz);
  decR= decomp_sz;

  if (uncompress(dec, &decR, raw+9, rawlen-9)!=Z_OK || (decR!=decomp_sz))
  {
    free(raw);
    *Rfreep= *Rcmd= (unsigned char*) strdup("");
    *Rcmdlen= 0;
    return ;
  }

  free(raw);
  *Rfreep= *Rcmd= dec;
  *Rcmdlen= decomp_sz;
  return ;
}


static uint32_t write_page
   (FILE *fO, uint16_t w, uint16_t h, uint8_t color,
              uint8_t *inp, int cmdlen)
{
  unsigned char TB[20];
  unsigned char *data;
  int compressed_len,j;

  // this sets cl=pl and then pl=0 if the compression
  // fails or doesn't make sense. otherwise pl isn't touched
  // and the uncompressed buffer is freed.
  data= optional_compress_commands(inp, &cmdlen, &compressed_len);

  j= 0;
                   // uncompressed command length, width, height, color
  j+= Oint4(TB+j, 4+ 2 + 2 + 1 + compressed_len);         // size
  j+= Oint4(TB+j, cmdlen);
  j+= Oint2(TB+j, w);
  j+= Oint2(TB+j, h);
  j+= Oint1(TB+j, color); 

  write_bytes(fO, TB, j);
  write_bytes(fO, data, compressed_len);

  if (data!=inp) free(data);
  return j+compressed_len;
}


static uint32_t copy_pages(FILE *fO, inp_t *inp, uint32_t offset, 
                       uint32_t rename, uint32_t *page_table)
{
  uint32_t i,j,N;

  for(i=0;i<inp->n_page;i++)
  {
    uint8_t *data; uint32_t size;
    uint16_t w, h; uint8_t color;
    uint8_t *freep;

    page_table[rename+i]= offset;
    read_page(inp, i, &w,&h,&color,&data, &size, &freep);
    for(j=0;j<size;)
    switch(data[j])
    {
                  // code x,y,w,h color
    case 'R': j+= 1+ 2+2+2+2+1; break;
                  // code x,y,no
    case 'B': j+= 1+ 2+2;
              N= dint4(data+j);
              if (N<inp->n_bmap)
                 Oint4(data+j, inp->bmap_rename[N]);
              j+= 4;
              break;
    case 'G': j+= 1+ 2+2;
              N= dint4(data+j);
              if (N<inp->n_gmap)
                Oint4(data+j, inp->gmap_rename[N]);
              j+= 4;
              break;
    }
    offset+= write_page(fO, w,h,color,data, size);
    free(freep);
  } 
  return offset;
}

static void combine_pages(FILE *fO, inp_t *inpA, inp_t *inpB)
{
  uint32_t *page_table;
  uint32_t n_page, SO_page, table_size, offset, SS_page;

  n_page= inpA->n_page+inpB->n_page;
  page_table= calloc(n_page, 4);

  seek_end(fO);
  SO_page= get_fpos(fO);

  table_size= 4*n_page;
  write_padding(fO,table_size+4);
  offset= SO_page + table_size + 4;
  offset= copy_pages(fO, inpA, offset, 0, page_table);
  offset= copy_pages(fO, inpB, offset, inpA->n_page, page_table);

  SS_page= offset - SO_page;
  
  set_fpos(fO, SO_page);
  write_table(fO, n_page, page_table);
  free(page_table);

  set_fpos(fO, HDR_PAGE);
  write_int4(fO, SS_page);
  write_int4(fO, SO_page);
}

static bmk_t *get_bmk(inp_t *inp)
{
  uint32_t sz;
  uint8_t *data;
  bmk_t *B;
  if (inp->SO_bmk==0) return NULL;

  set_fpos(inp->f, inp->SO_bmk);
  sz= read_int4(inp);
  data= malloc(sz);
  read_bytes(inp->f, data, sz);
  B= bmk_decode(data, sz);
  free(data);
  return B;
}

static void rename_bookmark_pages(bmk_t *B,int offset)
{
  bmk_t *C;
  B->page+= offset;
  if (B->children)
    for(C=B->children->next;C!=B->children;C=C->next)
       rename_bookmark_pages(C, offset);
}

static void combine_bookmarks(FILE *fO, inp_t *inpA, inp_t *inpB)
{
  bmk_t *bmkA, *bmkB;
  uint8_t *data;
  uint32_t size;
  uint32_t SS_bmk, SO_bmk;

    // last child of A, 'children' node of A
                // first child of B, last child of B
  bmk_t *lcA, *cA, *fcB, *lcB;
  bmkA= get_bmk(inpA);
  bmkB= get_bmk(inpB);
  if (bmkB) rename_bookmark_pages(bmkB, inpA->n_page);

  if (!bmkA || !bmkA->children) { bmkA= bmkB; goto done; }
  if (!bmkB || !bmkB->children) goto done; 

  fcB= bmkB->children->next;
  lcB= bmkB->children->prev;

  if (fcB==bmkB->children)  goto done; 

  cA= bmkA->children;
  lcA= bmkA->children->prev;

  lcA->next= fcB;
  cA->prev= lcB;

  fcB->prev= lcA;
  lcB->next= cA;

  unify_bmks(cA);

done:
  if (!bmkA || !bmkA->children || bmkA->children->next==bmkA->children)
    return ;
  data= bmk_encode(bmkA, &size);

  seek_end(fO);
  SO_bmk= get_fpos(fO);
  SS_bmk= size;

  write_bytes(fO, data, size);
  free(data);

  set_fpos(fO, HDR_BMK);
  write_int4(fO, SS_bmk);
  write_int4(fO, SO_bmk);
}

int main(int argc,char **argv)
{
  FILE  *fO;
  inp_t inpA, inpB;

  if (argc!=4) 
    exit(fprintf(stderr,"usage: %s input1.imd input2.imd output.imd\n",
                    argv[0]));
  inpA.f= open_force(argv[1],"rb"); inpA.name= argv[1];
  inpB.f= open_force(argv[2],"rb"); inpB.name= argv[2];

  fO= open_force(argv[3],"wb");

  write_padding(fO, MAGIC_SIZE);

  read_header_and_tables(&inpA);
  read_header_and_tables(&inpB);

  combine_bitmaps(fO,&inpA, &inpB);
  combine_graymaps(fO,&inpA, &inpB);
  combine_pages(fO,&inpA, &inpB);
  combine_bookmarks(fO,&inpA, &inpB);

  set_fpos(fO, 0);
  write_bytes(fO, MAGIC, MAGIC_SIZE);

  fclose(fO);
  return 0;
}


